#define USE_RINTERNALS

#include <R.h>
#include <Rinternals.h>
#include <Rmath.h>
#include <pthread.h>
#include "workers.h"
#include "pnmath.h"

#ifdef NEED_DUPLICATE_ATTRIB
static void (DUPLICATE_ATTRIB)(SEXP to, SEXP from) {
    SET_ATTRIB(to, duplicate(ATTRIB(from)));
    SET_OBJECT(to, OBJECT(from));
    IS_S4_OBJECT(from) ?  SET_S4_OBJECT(to) : UNSET_S4_OBJECT(to);
}
#endif

#define DUPLICATE_ATTRIB(to, from) do {\
  SEXP __from__ = (from); \
  if (ATTRIB(__from__) != R_NilValue) { \
    SEXP __to__ = (to); \
    (DUPLICATE_ATTRIB)(__to__, __from__);       \
  } \
} while (0)

#define _(String) (String)
#define R_MSG_NA  _("NaNs produced")
#define R_MSG_NONNUM_MATH _("Non-numeric argument to mathematical function")

#define incMod(i,n) (i = (++i == n) ? 0 : i)


/*
  Worker Management Utilities
*/

#define MAX_NUM_JOBS 128

static void signal_error(const char *msg)
{
    error(msg);
}

/*
  Data base for cutoffs determining the number of threads to use
*/

#define MAX_VNMATH 1000
#define VNNAME_MAX 256
static int nvnmath = 0;
static struct { char name[VNNAME_MAX]; int n2, n4, n8; } vnmath[MAX_VNMATH];

static int findVNmathIndex(const char *name) {
    int i;
    for (i = 0; i < nvnmath; i++)
	if (strncmp(name, vnmath[i].name, VNNAME_MAX) == 0)
	    return i;
    return -1;
}

static int vnmathEnter(const char *name, int n2, int n4, int n8)
{
    int idx = findVNmathIndex(name);
    if (idx < 0) {
	if (nvnmath < MAX_VNMATH) {
	    idx = nvnmath;
	    nvnmath++;
	}
	else error("Table is full.");
    }
    strncpy(vnmath[idx].name, name, VNNAME_MAX);
    vnmath[idx].n2 = n2;
    vnmath[idx].n4 = n4;
    vnmath[idx].n8 = n8;
    return idx;
}

SEXP do_vnmathEnter(SEXP sName, SEXP sN2, SEXP sN4, SEXP sN8)
{
    const char *name;
    int n2, n4, n8, idx;

    if (TYPEOF(sName) != STRSXP || LENGTH(sName) != 1)
	error("bad 'name' argument type");
    name = CHAR(STRING_ELT(sName, 0));
    n2 = asInteger(sN2);
    n4 = asInteger(sN4);
    n8 = asInteger(sN8);
    idx = vnmathEnter(name, n2, n4, n8);
    return ScalarInteger(idx);
}

/*
  Determining the processor count
*/

static int PMAX = 1;

/* For now this uses 1, 2, 4, or 8 processors. */
#define FIXUP_NUM_PROCS do { \
    int idx = -P; \
    if (n > vnmath[idx].n2 && vnmath[idx].n2 > 0 && PMAX >= 2) \
	if (n > vnmath[idx].n4 && vnmath[idx].n4 > 0 && PMAX >= 4) \
	    if (n > vnmath[idx].n8 && vnmath[idx].n8 > 0 && PMAX >= 8) \
		P = 8; \
	    else P = 4; \
	else P = 2; \
    else P = 1; \
    } while (0)


/*
  Math 1 Group
*/

typedef struct {
    int N, P, n;
    double *a, *y;
    int na;
    int naflag[MAX_NUM_JOBS];
    double (*f)(double);
} math1_work_data_t, *math1_work_data_p;

#define MATH1_LOOP(start, starta, end, call, naflag) \
    do { \
	for (i = (start), ia = (starta); \
	     i < (end); \
             incMod(ia, na), i++) { \
	ai = a[ia]; \
	if (ISNA (ai)) y[i] = NA_REAL; \
	else if (ISNAN(ai)) y[i] = R_NaN; \
	else { \
	    y[i] = (call); \
	    if (ISNAN(y[i])) (naflag) = 1; \
	} \
    } \
    } while(0)

static void math1_work(int k, void *data)
{
    math1_work_data_p mydata = (math1_work_data_p) data;
    int N = mydata->N;
    int P = mydata->P;
    int i, start, end, ia;
    double ai;
    double *a = mydata->a, *y = mydata->y;
    int na = mydata->na;
    double (*f)(double) = mydata->f;

    start = k * N;
    if (k == P - 1) end = mydata->n;
    else end = (k + 1) * N;
    mydata->naflag[k] = 0;

    MATH1_LOOP(start, start % na, end, f(ai), mydata->naflag[k]);
}

static SEXP math1(SEXP sa, double (*f)(double), SEXP lcall, int P)
{
    SEXP sy;
    int i, ia, n, na;
    double ai, *a, *y;
    int naflag;

    if (!isNumeric(sa))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa);

    if (na == 0) return(allocVector(REALSXP, 0));

    n = na;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); y = REAL(sy); naflag = 0;

    if (P <= 0) FIXUP_NUM_PROCS;
		    
    if (P <= 1)
	MATH1_LOOP(0, 0, n, f(ai), naflag);
    else {
	int N;
	math1_work_data_t data;

	if (P > n) P = n;
	if (P > MAX_NUM_JOBS) P = MAX_NUM_JOBS;
	N = n / P;
	if (N * P < n) N++;
	data.N = N;
	data.P = P;
	data.n = n;
	data.a = a; data.y = y;
	data.na = na;
	data.f = f;

	runWorkers(P, math1_work, &data);
	for (i = 0; i < P; i++) 
	    if (data.naflag[i]) naflag = 1;
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    DUPLICATE_ATTRIB(sy, sa);
    UNPROTECT(2);
    return sy;
} /* math1 */

int PRIMVAL(SEXP);

#ifdef isObject
# undef isObject
#endif
#define isObject(s)   (OBJECT(s) != 0)

#define MIGHT_NEED_DISPATCH(args) \
    ! (args != R_NilValue && ! isObject(CAR(args)) &&		\
       (CDR(args) == R_NilValue || ! isObject(CADR(args))))

#define checkArity(a,b) Rf_checkArityCall(a,b,call)
void Rf_checkArityCall(SEXP, SEXP, SEXP);

#ifdef isComplex
# undef isComplex
#endif
#define isComplex(s) (TYPEOF(s) == CPLXSXP)
static SEXP (*complex_math1)(SEXP, SEXP, SEXP, SEXP);

static SEXP (*do_math1)(SEXP, SEXP, SEXP, SEXP);
static SEXP (*do_math2)(SEXP, SEXP, SEXP, SEXP);
static SEXP (*do_math3)(SEXP, SEXP, SEXP, SEXP);
static SEXP (*do_math4)(SEXP, SEXP, SEXP, SEXP);

#define DEFMATH_V_SETUP(name)  \
    static int idx = -1;       \
    int p = asInteger(sP);     \
    if (p == 0) {					\
	if (idx < 0) idx = findVNmathIndex(#name);	\
	p = -idx;					\
    }

#define DEFMATH_1(name) \
    SEXP do_##name##_v(SEXP sa, SEXP sP) { \
	DEFMATH_V_SETUP(name) \
        return math1(sa, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa; \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (PRIMVAL(op) != 46 && \
            MIGHT_NEED_DISPATCH(args)) \
	    return do_math1(call, op, args, env); \
	if (isComplex(CAR(args))) \
	    return complex_math1(call, op, args, env);\
	sa = CAR(args); \
	return math1(sa, name, R_NilValue, -idx); }

DEFMATH_1(floor)
DEFMATH_1(ceil)
DEFMATH_1(sqrt)
DEFMATH_1(sign)

DEFMATH_1(exp)
DEFMATH_1(expm1)
DEFMATH_1(log1p)
DEFMATH_1(cos)
DEFMATH_1(sin)
DEFMATH_1(tan)
DEFMATH_1(acos)
DEFMATH_1(asin)

DEFMATH_1(cosh)
DEFMATH_1(sinh)
DEFMATH_1(tanh)
DEFMATH_1(acosh)
DEFMATH_1(asinh)
DEFMATH_1(atanh)

DEFMATH_1(lgammafn)
DEFMATH_1(gammafn)

DEFMATH_1(digamma)
DEFMATH_1(trigamma)

DEFMATH_1(gamma_cody)


/*
  Math 2 Group
*/

typedef struct {
    int N, P, n;
    double *a, *b, *y;
    int na, nb;
    int naflag[MAX_NUM_JOBS];
    double (*f)(double, double);
} math2_work_data_t, *math2_work_data_p;

typedef struct {
    int N, P, n;
    double *a, *b, *y;
    int na, nb, i_1;
    int naflag[MAX_NUM_JOBS];
    double (*f)(double, double, int);
} math2_1_work_data_t, *math2_1_work_data_p;

typedef struct {
    int N, P, n;
    double *a, *b, *y;
    int na, nb, i_1, i_2;
    int naflag[MAX_NUM_JOBS];
    double (*f)(double, double, int, int);
} math2_2_work_data_t, *math2_2_work_data_p;

#define MATH2_LOOP(start, starta, startb, end, call, naflag) \
    do { \
	for (i = (start), ia = (starta), ib = (startb); \
	     i < (end); \
             incMod(ia, na), incMod(ib, nb), i++) { \
	ai = a[ia]; \
	bi = b[ib]; \
	if (ISNA (ai) || ISNA (bi)) y[i] = NA_REAL; \
	else if (ISNAN(ai) || ISNAN(bi)) y[i] = R_NaN; \
	else { \
	    y[i] = (call); \
	    if (ISNAN(y[i])) (naflag) = 1; \
	} \
    } \
    } while(0)

static void math2_work(int k, void *data)
{
    math2_work_data_p mydata = (math2_work_data_p) data;
    int N = mydata->N;
    int P = mydata->P;
    int i, start, end, ia, ib;
    double ai, bi;
    double *a = mydata->a, *b = mydata->b, *y = mydata->y;
    int na = mydata->na, nb = mydata->nb;
    double (*f)(double, double) = mydata->f;

    start = k * N;
    if (k == P - 1) end = mydata->n;
    else end = (k + 1) * N;
    mydata->naflag[k] = 0;

    MATH2_LOOP(start, start % na, start % nb, end,
	       f(ai, bi), mydata->naflag[k]);
}

static void math2_1_work(int k, void *data)
{
    math2_1_work_data_p mydata = (math2_1_work_data_p) data;
    int N = mydata->N;
    int P = mydata->P;
    int i, start, end, ia, ib;
    double ai, bi;
    double *a = mydata->a, *b = mydata->b, *y = mydata->y;
    int na = mydata->na, nb = mydata->nb;
    int i_1 = mydata->i_1;
    double (*f)(double, double, int) = mydata->f;

    start = k * N;
    if (k == P - 1) end = mydata->n;
    else end = (k + 1) * N;
    mydata->naflag[k] = 0;

    MATH2_LOOP(start, start % na, start % nb, end,
	       f(ai, bi, i_1), mydata->naflag[k]);
}

static void math2_2_work(int k, void *data)
{
    math2_2_work_data_p mydata = (math2_2_work_data_p) data;
    int N = mydata->N;
    int P = mydata->P;
    int i, start, end, ia, ib;
    double ai, bi;
    double *a = mydata->a, *b = mydata->b, *y = mydata->y;
    int na = mydata->na, nb = mydata->nb;
    int i_1 = mydata->i_1, i_2 = mydata->i_2;
    double (*f)(double, double, int, int) = mydata->f;

    start = k * N;
    if (k == P - 1) end = mydata->n;
    else end = (k + 1) * N;
    mydata->naflag[k] = 0;

    MATH2_LOOP(start, start % na, start % nb, end,
	       f(ai, bi, i_1, i_2), mydata->naflag[k]);
}

static SEXP math2(SEXP sa, SEXP sb,
		  double (*f)(double, double), SEXP lcall,
		  int P)
{
    SEXP sy;
    int i, ia, ib, n, na, nb;
    double ai, bi, *a, *b, *y;
    int naflag;

    if (!isNumeric(sa) || !isNumeric(sb))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa); nb = LENGTH(sb);

    if ((na == 0) || (nb == 0)) return(allocVector(REALSXP, 0));

    n = na; if (n < nb) n = nb;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sb = coerceVector(sb, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); b = REAL(sb); y = REAL(sy); naflag = 0;

    if (P <= 0) FIXUP_NUM_PROCS;
		    
    if (P <= 1)
	MATH2_LOOP(0, 0, 0, n, f(ai, bi), naflag);
    else {
	int N;
	math2_work_data_t data;

	if (P > n) P = n;
	if (P > MAX_NUM_JOBS) P = MAX_NUM_JOBS;
	N = n / P;
	if (N * P < n) N++;
	data.N = N;
	data.P = P;
	data.n = n;
	data.a = a; data.b = b; data.y = y;
	data.na = na; data.nb = nb;
	data.f = f;

	runWorkers(P, math2_work, &data);
	for (i = 0; i < P; i++) 
	    if (data.naflag[i]) naflag = 1;
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    if (n == na) { DUPLICATE_ATTRIB(sy, sa); }
    else if (n == nb) { DUPLICATE_ATTRIB(sy, sb); }
    UNPROTECT(3);
    return sy;
} /* math2 */

static SEXP math2_1(SEXP sa, SEXP sb, SEXP sI,
		    double (*f)(double, double, int), SEXP lcall,
		    int P)
{
    SEXP sy;
    int i, ia, ib, n, na, nb;
    double ai, bi, *a, *b, *y;
    int i_1;
    int naflag;

    if (!isNumeric(sa) || !isNumeric(sb))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa); nb = LENGTH(sb);

    if ((na == 0) || (nb == 0)) return(allocVector(REALSXP, 0));

    n = na; if (n < nb) n = nb;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sb = coerceVector(sb, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); b = REAL(sb); y = REAL(sy); naflag = 0;
    i_1 = asInteger(sI);

    if (P <= 0) FIXUP_NUM_PROCS;
		    
    if (P <= 1)
	MATH2_LOOP(0, 0, 0, n, f(ai, bi, i_1), naflag);
    else {
	int N;
	math2_1_work_data_t data;

	if (P > n) P = n;
	if (P > MAX_NUM_JOBS) P = MAX_NUM_JOBS;
	N = n / P;
	if (N * P < n) N++;
	data.N = N;
	data.P = P;
	data.n = n;
	data.a = a; data.b = b; data.y = y;
	data.na = na; data.nb = nb;
	data.i_1 = i_1;
	data.f = f;

	runWorkers(P, math2_1_work, &data);
	for (i = 0; i < P; i++) 
	    if (data.naflag[i]) naflag = 1;
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    if (n == na) { DUPLICATE_ATTRIB(sy, sa); }
    else if (n == nb) { DUPLICATE_ATTRIB(sy, sb); }
    UNPROTECT(3);
    return sy;
} /* math2_1 */

static SEXP math2_2(SEXP sa, SEXP sb, SEXP sI, SEXP sJ,
		    double (*f)(double, double, int, int), SEXP lcall,
		    int P)
{
    SEXP sy;
    int i, ia, ib, n, na, nb;
    double ai, bi, *a, *b, *y;
    int i_1,i_2;
    int naflag;

    if (!isNumeric(sa) || !isNumeric(sb))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa); nb = LENGTH(sb);

    if ((na == 0) || (nb == 0)) return(allocVector(REALSXP, 0));

    n = na; if (n < nb) n = nb;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sb = coerceVector(sb, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); b = REAL(sb); y = REAL(sy); naflag = 0;
    i_1 = asInteger(sI);
    i_2 = asInteger(sJ);

    if (P <= 0) FIXUP_NUM_PROCS;
		    
    if (P <= 1)
	MATH2_LOOP(0, 0, 0, n, f(ai, bi, i_1, i_2), naflag);
    else {
	int N;
	math2_2_work_data_t data;

	if (P > n) P = n;
	if (P > MAX_NUM_JOBS) P = MAX_NUM_JOBS;
	N = n / P;
	if (N * P < n) N++;
	data.N = N;
	data.P = P;
	data.n = n;
	data.a = a; data.b = b; data.y = y;
	data.na = na; data.nb = nb;
	data.i_1 = i_1; data.i_2 = i_2;
	data.f = f;

	runWorkers(P, math2_2_work, &data);
	for (i = 0; i < P; i++) 
	    if (data.naflag[i]) naflag = 1;
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    if (n == na) { DUPLICATE_ATTRIB(sy, sa); }
    else if (n == nb) { DUPLICATE_ATTRIB(sy, sb); }
    UNPROTECT(3);
    return sy;
} /* math2_2 */

static SEXP (*complex_math2)(SEXP, SEXP, SEXP, SEXP);

#define DEFMATH_2(name)	\
    SEXP do_##name##_v(SEXP sa, SEXP sb, SEXP sP) { \
	DEFMATH_V_SETUP(name) \
        return math2(sa, sb, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa, sb; \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (isComplex(CAR(args)) || \
            (PRIMVAL(op) == 0 && isComplex(CADR(args)))) \
            return complex_math2(call, op, args, env); \
	sa = CAR(args); \
	sb = CADR(args); \
        return math2(sa, sb, name, R_NilValue, -idx); }

#define DEFMATH_2_1(name) \
    SEXP do_##name##_v(SEXP sa, SEXP sb, SEXP sI, SEXP sP) { \
	DEFMATH_V_SETUP(name) \
        return math2_1(sa, sb, sI, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa, sb, sI; \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (isComplex(CAR(args)) || \
            (PRIMVAL(op) == 0 && isComplex(CADR(args)))) \
            return complex_math2(call, op, args, env); \
	sa = CAR(args); \
	sb = CADR(args); \
	sI = CADDR(args); \
        return math2_1(sa, sb, sI, name, R_NilValue, -idx); }

#define DEFMATH_2_2(name) \
    SEXP do_##name##_v(SEXP sa, SEXP sb, SEXP sI, SEXP sJ, SEXP sP) { \
	DEFMATH_V_SETUP(name) \
        return math2_2(sa, sb, sI, sJ, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa, sb, sI, sJ;					  \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (isComplex(CAR(args)) || \
            (PRIMVAL(op) == 0 && isComplex(CADR(args)))) \
            return complex_math2(call, op, args, env); \
	sa = CAR(args); \
	sb = CADR(args); \
	sI = CADDR(args); \
	sJ = CADDDR(args); \
        return math2_2(sa, sb, sI, sJ, name, R_NilValue, -idx); }

DEFMATH_2(lbeta)
DEFMATH_2(beta)
DEFMATH_2(lchoose)
DEFMATH_2(choose)

DEFMATH_2_1(dchisq)
DEFMATH_2_2(pchisq)
DEFMATH_2_2(qchisq)

DEFMATH_2_1(dexp)
DEFMATH_2_2(pexp)
DEFMATH_2_2(qexp)

DEFMATH_2_1(dgeom)
DEFMATH_2_2(pgeom)
DEFMATH_2_2(qgeom)

DEFMATH_2_1(dpois)
DEFMATH_2_2(ppois)
DEFMATH_2_2(qpois)

DEFMATH_2_1(dt)
DEFMATH_2_2(pt)
DEFMATH_2_2(qt)

#ifdef NOT_THREAD_SAFE
DEFMATH_2_1(dsignrank)
DEFMATH_2_2(psignrank)
DEFMATH_2_2(qsignrank)

DEFMATH_2(bessel_j)
DEFMATH_2(bessel_y)
#endif
DEFMATH_2(psigamma)


/*
  Math 3 Group
*/

typedef struct {
    int N, P, n;
    double *a, *b, *c, *y;
    int na, nb, nc, i_1;
    int naflag[MAX_NUM_JOBS];
    double (*f)(double, double, double, int);
} math3_1_work_data_t, *math3_1_work_data_p;

typedef struct {
    int N, P, n;
    double *a, *b, *c, *y;
    int na, nb, nc, i_1, i_2;
    int naflag[MAX_NUM_JOBS];
    double (*f)(double, double, double, int, int);
} math3_2_work_data_t, *math3_2_work_data_p;

#define MATH3_LOOP(start, starta, startb, startc, end, call, naflag) \
    do { \
	for (i = (start), ia = (starta), ib = (startb), ic = (startc); \
	     i < (end); \
	 incMod(ia, na), incMod(ib, nb), incMod(ic, nc), i++) { \
	ai = a[ia]; \
	bi = b[ib]; \
	ci = c[ic]; \
	if (ISNA (ai) || ISNA (bi)|| ISNA (ci)) y[i] = NA_REAL; \
	else if (ISNAN(ai) || ISNAN(bi)|| ISNAN(ci)) y[i] = R_NaN; \
	else { \
	    y[i] = (call); \
	    if (ISNAN(y[i])) (naflag) = 1; \
	} \
    } \
    } while(0)

static void math3_1_work(int k, void *data)
{
    math3_1_work_data_p mydata = (math3_1_work_data_p) data;
    int N = mydata->N;
    int P = mydata->P;
    int i, start, end, ia, ib, ic;
    double ai, bi, ci;
    double *a = mydata->a, *b = mydata->b, *c = mydata->c, *y = mydata->y;
    int na = mydata->na, nb = mydata->nb, nc = mydata->nc;
    int i_1 = mydata->i_1;
    double (*f)(double, double, double, int) = mydata->f;

    start = k * N;
    if (k == P - 1) end = mydata->n;
    else end = (k + 1) * N;
    mydata->naflag[k] = 0;

    MATH3_LOOP(start, start % na, start % nb, start % nc, end,
	       f(ai, bi, ci, i_1), mydata->naflag[k]);
}

static void math3_2_work(int k, void *data)
{
    math3_2_work_data_p mydata = (math3_2_work_data_p) data;
    int N = mydata->N;
    int P = mydata->P;
    int i, start, end, ia, ib, ic;
    double ai, bi, ci;
    double *a = mydata->a, *b = mydata->b, *c = mydata->c, *y = mydata->y;
    int na = mydata->na, nb = mydata->nb, nc = mydata->nc;
    int i_1 = mydata->i_1, i_2 = mydata->i_2;
    double (*f)(double, double, double, int, int) = mydata->f;

    start = k * N;
    if (k == P - 1) end = mydata->n;
    else end = (k + 1) * N;
    mydata->naflag[k] = 0;

    MATH3_LOOP(start, start % na, start % nb, start % nc, end,
	       f(ai, bi, ci, i_1, i_2), mydata->naflag[k]);
}

static SEXP math3_1(SEXP sa, SEXP sb, SEXP sc, SEXP sI,
		    double (*f)(double, double, double, int), SEXP lcall,
		    int P)
{
    SEXP sy;
    int i, ia, ib, ic, n, na, nb, nc;
    double ai, bi, ci, *a, *b, *c, *y;
    int i_1;
    int naflag;

    if (!isNumeric(sa) || !isNumeric(sb) || !isNumeric(sc))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa); nb = LENGTH(sb); nc = LENGTH(sc);

    if ((na == 0) || (nb == 0) || (nc == 0)) return(allocVector(REALSXP, 0));

    n = na; if (n < nb) n = nb; if (n < nc) n = nc;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sb = coerceVector(sb, REALSXP));
    PROTECT(sc = coerceVector(sc, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); b = REAL(sb); c = REAL(sc); y = REAL(sy); naflag = 0;
    i_1 = asInteger(sI);

    if (P <= 0) FIXUP_NUM_PROCS;
		    
    if (P <= 1)
	MATH3_LOOP(0, 0, 0, 0, n, f(ai, bi, ci, i_1), naflag);
    else {
	int N;
	math3_1_work_data_t data;

	if (P > n) P = n;
	if (P > MAX_NUM_JOBS) P = MAX_NUM_JOBS;
	N = n / P;
	if (N * P < n) N++;
	data.N = N;
	data.P = P;
	data.n = n;
	data.a = a; data.b = b;	data.c = c; data.y = y;
	data.na = na; data.nb = nb; data.nc = nc;
	data.i_1 = i_1;
	data.f = f;

	runWorkers(P, math3_1_work, &data);
	for (i = 0; i < P; i++) 
	    if (data.naflag[i]) naflag = 1;
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    if (n == na) { DUPLICATE_ATTRIB(sy, sa); }
    else if (n == nb) { DUPLICATE_ATTRIB(sy, sb); }
    else if (n == nc) { DUPLICATE_ATTRIB(sy, sc); }
    UNPROTECT(4);
    return sy;
} /* math3_1 */

static SEXP math3_2(SEXP sa, SEXP sb, SEXP sc, SEXP sI, SEXP sJ,
		    double (*f)(double, double, double, int, int), SEXP lcall,
		    int P)
{
    SEXP sy;
    int i, ia, ib, ic, n, na, nb, nc;
    double ai, bi, ci, *a, *b, *c, *y;
    int i_1,i_2;
    int naflag;

    if (!isNumeric(sa) || !isNumeric(sb) || !isNumeric(sc))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa); nb = LENGTH(sb); nc = LENGTH(sc);

    if ((na == 0) || (nb == 0) || (nc == 0)) return(allocVector(REALSXP, 0));

    n = na; if (n < nb) n = nb; if (n < nc) n = nc;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sb = coerceVector(sb, REALSXP));
    PROTECT(sc = coerceVector(sc, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); b = REAL(sb); c = REAL(sc); y = REAL(sy); naflag = 0;
    i_1 = asInteger(sI);
    i_2 = asInteger(sJ);

    if (P <= 0) FIXUP_NUM_PROCS;
		    
    if (P <= 1)
	MATH3_LOOP(0, 0, 0, 0, n, f(ai, bi, ci, i_1, i_2), naflag);
    else {
	int N;
	math3_2_work_data_t data;

	if (P > n) P = n;
	if (P > MAX_NUM_JOBS) P = MAX_NUM_JOBS;
	N = n / P;
	if (N * P < n) N++;
	data.N = N;
	data.P = P;
	data.n = n;
	data.a = a; data.b = b;	data.c = c; data.y = y;
	data.na = na; data.nb = nb; data.nc = nc;
	data.i_1 = i_1; data.i_2 = i_2;
	data.f = f;

	runWorkers(P, math3_2_work, &data);
	for (i = 0; i < P; i++) 
	    if (data.naflag[i]) naflag = 1;
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    if (n == na) { DUPLICATE_ATTRIB(sy, sa); }
    else if (n == nb) { DUPLICATE_ATTRIB(sy, sb); }
    else if (n == nc) { DUPLICATE_ATTRIB(sy, sc); }
    UNPROTECT(4);
    return sy;
} /* math3_2 */

#define DEFMATH_3_1(name) \
    SEXP do_##name##_v(SEXP sa, SEXP sb, SEXP sc, SEXP sI, SEXP sP) {	\
	DEFMATH_V_SETUP(name) \
        return math3_1(sa, sb, sc, sI, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa, sb, sc, sI;					  \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (isComplex(CAR(args)) || \
            (PRIMVAL(op) == 0 && isComplex(CADR(args)))) \
            return complex_math2(call, op, args, env); \
	sa = CAR(args); \
	sb = CADR(args); \
	sc = CADDR(args); \
	sI = CADDDR(args); \
        return math3_1(sa, sb, sc, sI, name, R_NilValue, -idx); }
#define DEFMATH_3_2(name) \
    SEXP do_##name##_v(SEXP sa, SEXP sb, SEXP sc, SEXP sI, SEXP sJ, \
		       SEXP sP) {				    \
	DEFMATH_V_SETUP(name) \
        return math3_2(sa, sb, sc, sI, sJ, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa, sb, sc, sI, sJ;					  \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (isComplex(CAR(args)) || \
            (PRIMVAL(op) == 0 && isComplex(CADR(args)))) \
            return complex_math2(call, op, args, env); \
	sa = CAR(args); \
	sb = CADR(args); \
	sc = CADDR(args); \
	sI = CADDDR(args); \
	sJ = CAD4R(args); \
        return math3_2(sa, sb, sc, sI, sJ, name, R_NilValue, -idx); }

DEFMATH_3_1(dbeta)
DEFMATH_3_2(pbeta)
DEFMATH_3_2(qbeta)

DEFMATH_3_1(dbinom)
DEFMATH_3_2(pbinom)
DEFMATH_3_2(qbinom)

DEFMATH_3_1(dcauchy)
DEFMATH_3_2(pcauchy)
DEFMATH_3_2(qcauchy)

DEFMATH_3_1(df)
DEFMATH_3_2(pf)
DEFMATH_3_2(qf)

DEFMATH_3_1(dgamma)
DEFMATH_3_2(pgamma)
DEFMATH_3_2(qgamma)

DEFMATH_3_1(dlnorm)
DEFMATH_3_2(plnorm)
DEFMATH_3_2(qlnorm)

DEFMATH_3_1(dlogis)
DEFMATH_3_2(plogis)
DEFMATH_3_2(qlogis)

DEFMATH_3_1(dnbinom)
DEFMATH_3_2(pnbinom)
DEFMATH_3_2(qnbinom)

DEFMATH_3_1(dnorm)
DEFMATH_3_2(pnorm)
DEFMATH_3_2(qnorm)

DEFMATH_3_1(dunif)
DEFMATH_3_2(punif)
DEFMATH_3_2(qunif)

DEFMATH_3_1(dweibull)
DEFMATH_3_2(pweibull)
DEFMATH_3_2(qweibull)

DEFMATH_3_1(dnchisq)
DEFMATH_3_2(pnchisq)
DEFMATH_3_2(qnchisq)

DEFMATH_3_1(dnt)
DEFMATH_3_2(pnt)
DEFMATH_3_2(qnt)

#ifdef NOT_THREAD_SAFE
DEFMATH_3_1(dwilcox)
DEFMATH_3_2(pwilcox)
DEFMATH_3_2(qwilcox)

DEFMATH3(bessel_i)
DEFMATH3(bessel_k)
#endif

/*
  Math 4 Group
*/

typedef struct {
    int N, P, n;
    double *a, *b, *c, *d, *y;
    int na, nb, nc, nd, i_1;
    int naflag[MAX_NUM_JOBS];
    double (*f)(double, double, double, double, int);
} math4_1_work_data_t, *math4_1_work_data_p;

typedef struct {
    int N, P, n;
    double *a, *b, *c, *d, *y;
    int na, nb, nc, nd, i_1, i_2;
    int naflag[MAX_NUM_JOBS];
    double (*f)(double, double, double, double, int, int);
} math4_2_work_data_t, *math4_2_work_data_p;

#define MATH4_LOOP(start, starta, startb, startc, startd, end, call, naflag) \
    do { \
	for (i = (start), ia = (starta), ib = (startb), ic = (startc), \
		 id = (startd);	\
	     i < (end); \
	     incMod(ia, na), incMod(ib, nb), incMod(ic, nc), incMod(id, nd), \
		 i++) { \
	ai = a[ia]; \
	bi = b[ib]; \
	ci = c[ic]; \
	di = d[id]; \
	if (ISNA (ai) || ISNA (bi)|| ISNA (ci) || ISNA(di)) y[i] = NA_REAL; \
	else if (ISNAN(ai) || ISNAN(bi)|| ISNAN(ci) || ISNAN(di)) \
	    y[i] = R_NaN; \
	else { \
	    y[i] = (call); \
	    if (ISNAN(y[i])) (naflag) = 1; \
	} \
    } \
    } while(0)

static void math4_1_work(int k, void *data)
{
    math4_1_work_data_p mydata = (math4_1_work_data_p) data;
    int N = mydata->N;
    int P = mydata->P;
    int i, start, end, ia, ib, ic, id;
    double ai, bi, ci, di;
    double *a = mydata->a, *b = mydata->b, *c = mydata->c, *d = mydata->d;
    double *y = mydata->y;
    int na = mydata->na, nb = mydata->nb, nc = mydata->nc, nd = mydata->nd;
    int i_1 = mydata->i_1;
    double (*f)(double, double, double, double, int) = mydata->f;

    start = k * N;
    if (k == P - 1) end = mydata->n;
    else end = (k + 1) * N;
    mydata->naflag[k] = 0;

    MATH4_LOOP(start, start % na, start % nb, start % nc, start % nd, end,
	       f(ai, bi, ci, di, i_1), mydata->naflag[k]);
}

static void math4_2_work(int k, void *data)
{
    math4_2_work_data_p mydata = (math4_2_work_data_p) data;
    int N = mydata->N;
    int P = mydata->P;
    int i, start, end, ia, ib, ic, id;
    double ai, bi, ci, di;
    double *a = mydata->a, *b = mydata->b, *c = mydata->c, *d = mydata->d;
    double *y = mydata->y;
    int na = mydata->na, nb = mydata->nb, nc = mydata->nc, nd = mydata->nd;
    int i_1 = mydata->i_1, i_2 = mydata->i_2;
    double (*f)(double, double, double, double, int, int) = mydata->f;

    start = k * N;
    if (k == P - 1) end = mydata->n;
    else end = (k + 1) * N;
    mydata->naflag[k] = 0;

    MATH4_LOOP(start, start % na, start % nb, start % nc, start % nd, end,
	       f(ai, bi, ci, di, i_1, i_2), mydata->naflag[k]);
}

static SEXP math4_1(SEXP sa, SEXP sb, SEXP sc, SEXP sd, SEXP sI,
		    double (*f)(double, double, double, double, int),
		    SEXP lcall, int P)
{
    SEXP sy;
    int i, ia, ib, ic, id, n, na, nb, nc, nd;
    double ai, bi, ci, di, *a, *b, *c, *d, *y;
    int i_1;
    int naflag;

    if (!isNumeric(sa) || !isNumeric(sb) || !isNumeric(sc) || !isNumeric(sd))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa); nb = LENGTH(sb); nc = LENGTH(sc); nd = LENGTH(sd);

    if ((na == 0) || (nb == 0) || (nc == 0) || (nd == 0))
	return(allocVector(REALSXP, 0));

    n = na; if (n < nb) n = nb; if (n < nc) n = nc; if (n < nd) n = nd;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sb = coerceVector(sb, REALSXP));
    PROTECT(sc = coerceVector(sc, REALSXP));
    PROTECT(sd = coerceVector(sd, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); b = REAL(sb); c = REAL(sc); d = REAL(sd); y = REAL(sy);
    naflag = 0;
    i_1 = asInteger(sI);

    if (P <= 0) FIXUP_NUM_PROCS;
		    
    if (P <= 1)
	MATH4_LOOP(0, 0, 0, 0, 0, n, f(ai, bi, ci, di, i_1), naflag);
    else {
	int N;
	math4_1_work_data_t data;

	if (P > n) P = n;
	if (P > MAX_NUM_JOBS) P = MAX_NUM_JOBS;
	N = n / P;
	if (N * P < n) N++;
	data.N = N;
	data.P = P;
	data.n = n;
	data.a = a; data.b = b;	data.c = c; data.d = d; data.y = y;
	data.na = na; data.nb = nb; data.nc = nc; data.nd = nd;
	data.i_1 = i_1;
	data.f = f;

	runWorkers(P, math4_1_work, &data);
	for (i = 0; i < P; i++) 
	    if (data.naflag[i]) naflag = 1;
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    if (n == na) { DUPLICATE_ATTRIB(sy, sa); }
    else if (n == nb) { DUPLICATE_ATTRIB(sy, sb); }
    else if (n == nc) { DUPLICATE_ATTRIB(sy, sc); }
    else if (n == nd) { DUPLICATE_ATTRIB(sy, sd); }
    UNPROTECT(5);
    return sy;
} /* math4_1 */

static SEXP math4_2(SEXP sa, SEXP sb, SEXP sc, SEXP sd, SEXP sI, SEXP sJ,
		    double (*f)(double, double, double, double, int, int),
		    SEXP lcall, int P)
{
    SEXP sy;
    int i, ia, ib, ic, id, n, na, nb, nc, nd;
    double ai, bi, ci, di, *a, *b, *c, *d, *y;
    int i_1,i_2;
    int naflag;

    if (!isNumeric(sa) || !isNumeric(sb) || !isNumeric(sc) || !isNumeric(sd))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa); nb = LENGTH(sb); nc = LENGTH(sc); nd = LENGTH(sd);

    if ((na == 0) || (nb == 0) || (nc == 0) || (nd == 0))
	return(allocVector(REALSXP, 0));

    n = na; if (n < nb) n = nb; if (n < nc) n = nc; if (n < nd) n = nd;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sb = coerceVector(sb, REALSXP));
    PROTECT(sc = coerceVector(sc, REALSXP));
    PROTECT(sd = coerceVector(sd, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); b = REAL(sb); c = REAL(sc); d = REAL(sd); y = REAL(sy);
    naflag = 0;
    i_1 = asInteger(sI);
    i_2 = asInteger(sJ);

    if (P <= 0) FIXUP_NUM_PROCS;
		    
    if (P <= 1)
	MATH4_LOOP(0, 0, 0, 0, 0, n, f(ai, bi, ci, di, i_1, i_2), naflag);
    else {
	int N;
	math4_2_work_data_t data;

	if (P > n) P = n;
	if (P > MAX_NUM_JOBS) P = MAX_NUM_JOBS;
	N = n / P;
	if (N * P < n) N++;
	data.N = N;
	data.P = P;
	data.n = n;
	data.a = a; data.b = b;	data.c = c; data.d = d; data.y = y;
	data.na = na; data.nb = nb; data.nc = nc; data.nd = nd;
	data.i_1 = i_1; data.i_2 = i_2;
	data.f = f;

	runWorkers(P, math4_2_work, &data);
	for (i = 0; i < P; i++) 
	    if (data.naflag[i]) naflag = 1;
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    if (n == na) { DUPLICATE_ATTRIB(sy, sa); }
    else if (n == nb) { DUPLICATE_ATTRIB(sy, sb); }
    else if (n == nc) { DUPLICATE_ATTRIB(sy, sc); }
    else if (n == nd) { DUPLICATE_ATTRIB(sy, sd); }
    UNPROTECT(5);
    return sy;
} /* math4_2 */

#define CAD5R(e)CAR(CDR(CDR(CDR(CDR(CDR(e))))))

#define DEFMATH_4_1(name) \
    SEXP do_##name##_v(SEXP sa, SEXP sb, SEXP sc, SEXP sd, SEXP sI, \
		       SEXP sP) {				    \
	DEFMATH_V_SETUP(name) \
        return math4_1(sa, sb, sc, sd, sI, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa, sb, sc, sd, sI; \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (isComplex(CAR(args)) || \
            (PRIMVAL(op) == 0 && isComplex(CADR(args)))) \
            return complex_math2(call, op, args, env); \
	sa = CAR(args); \
	sb = CADR(args); \
	sc = CADDR(args); \
	sd = CADDDR(args); \
	sI = CAD4R(args); \
        return math4_1(sa, sb, sc, sd, sI, name, R_NilValue, -idx); }
#define DEFMATH_4_2(name) \
    SEXP do_##name##_v(SEXP sa, SEXP sb, SEXP sc, SEXP sd, SEXP sI, SEXP sJ, \
		       SEXP sP) { \
	DEFMATH_V_SETUP(name) \
        return math4_2(sa, sb, sc, sd, sI, sJ, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa, sb, sc, sd, sI, sJ;				  \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (isComplex(CAR(args)) || \
            (PRIMVAL(op) == 0 && isComplex(CADR(args)))) \
            return complex_math2(call, op, args, env); \
	sa = CAR(args); \
	sb = CADR(args); \
	sc = CADDR(args); \
	sd = CADDDR(args); \
	sI = CAD4R(args); \
	sJ = CAD5R(args); \
        return math4_2(sa, sb, sc, sd, sI, sJ, name, R_NilValue, -idx); }

DEFMATH_4_1(dhyper)
DEFMATH_4_2(phyper)
DEFMATH_4_2(qhyper)

DEFMATH_4_1(dnbeta)
DEFMATH_4_2(pnbeta)
DEFMATH_4_2(qnbeta)

DEFMATH_4_1(dnf)
DEFMATH_4_2(pnf)
DEFMATH_4_2(qnf)

#ifdef UNIMP
DEFMATH_4_1(dtukey)
#endif
DEFMATH_4_2(ptukey)
DEFMATH_4_2(qtukey)


/*
  Replacing Builtin Functions.
 */

typedef SEXP (*CCODE)();
CCODE (PRIMFUN)(SEXP);
void (SET_PRIMFUN)(SEXP, CCODE);

CCODE R_get_arith_function(int);

static int getNumCPU(void);

SEXP do_initRpartest()
{
    PMAX = getNumCPU();
    if (PMAX < 0) PMAX = 1;
    if (PMAX > 1) initWorkers(PMAX, signal_error);

    do_math1 = R_get_arith_function(1);
    do_math2 = R_get_arith_function(2);
    do_math3 = R_get_arith_function(3);
    do_math4 = R_get_arith_function(4);
    complex_math1 = R_get_arith_function(11);
    complex_math2 = R_get_arith_function(12);
    return R_NilValue;
}

#ifdef WIN32
# include <windows.h>
#else
# include <unistd.h> /* for sysconf stuff */
# include <sys/types.h>
# include <sys/sysctl.h> /* for sysctl stuff */
#endif

static int getNumCPU(void)
{
    int numCPU = -1;
#if defined(WIN32)
    SYSTEM_INFO info;
    info.dwNumberOfProcessors = 0;
    GetSystemInfo (&info);
    numCPU = info.dwNumberOfProcessors;
#elif defined(HW_NCPU) /* BSD, e.g. Mac OS X */
    int mib[2];
    size_t len;

    mib[0] = CTL_HW;
    mib[1] = HW_NCPU;
    len = sizeof(numCPU);
    sysctl(mib, 2, &numCPU, &len, NULL, 0);
#elif defined(_SC_NPROCESSORS_CONF) /* Linux; maybe some others */
    numCPU = sysconf(_SC_NPROCESSORS_CONF);
#endif
    return numCPU;
}

SEXP do_NumProcs()
{
    return ScalarInteger(getNumCPU());
}

SEXP do_GetPMAX(void)
{
    return ScalarInteger(PMAX);
}

#define PMAX_SANITY_LIMIT MAX_NUM_JOBS

SEXP do_SetPMAX(SEXP sP)
{
    int P = asInteger(sP);
    if (P > 0 && P < PMAX_SANITY_LIMIT) {
	PMAX = P;
	initWorkers(PMAX, signal_error);
    }
    return ScalarInteger(PMAX);
}

static void replaceBuiltin(const char *name, CCODE oldfun, CCODE newfun)
{
    SEXP v, symbol = install(name);
    
    if ((v = INTERNAL(symbol)) != R_NilValue && TYPEOF(v) == BUILTINSXP
	&& PRIMFUN(v) == oldfun)
	SET_PRIMFUN(v, newfun);
    if ((v = SYMVALUE(symbol)) != R_UnboundValue && TYPEOF(v) == BUILTINSXP
	&& PRIMFUN(v) == oldfun)
	SET_PRIMFUN(v, newfun);
}
    
SEXP do_replaceBuiltins(void)
{
    /* Math 1 Group */
    replaceBuiltin("exp", do_math1, do_exp_i);

    /*replaceBuiltin("floor", do_math1, do_floor_i);
      replaceBuiltin("ceil", do_math1, do_ceil_i);*/
    replaceBuiltin("sqrt", do_math1, do_sqrt_i);
    /*replaceBuiltin("sign", do_math1, do_sign_i);*/

    replaceBuiltin("exp", do_math1, do_exp_i);
    replaceBuiltin("expm1", do_math1, do_expm1_i);
    replaceBuiltin("log1p", do_math1, do_log1p_i);
    replaceBuiltin("cos", do_math1, do_cos_i);
    replaceBuiltin("sin", do_math1, do_sin_i);
    replaceBuiltin("tan", do_math1, do_tan_i);
    replaceBuiltin("acos", do_math1, do_acos_i);
    replaceBuiltin("asin", do_math1, do_asin_i);

    replaceBuiltin("cosh", do_math1, do_cosh_i);
    replaceBuiltin("sinh", do_math1, do_sinh_i);
    replaceBuiltin("tanh", do_math1, do_tanh_i);
    replaceBuiltin("acosh", do_math1, do_acosh_i);
    replaceBuiltin("asinh", do_math1, do_asinh_i);
    replaceBuiltin("atanh", do_math1, do_atanh_i);

    replaceBuiltin("lgammafn", do_math1, do_lgammafn_i);
    replaceBuiltin("gammafn", do_math1, do_gammafn_i);

    replaceBuiltin("digamma", do_math1, do_digamma_i);
    replaceBuiltin("trigamma", do_math1, do_trigamma_i);

    replaceBuiltin("gamma_cody", do_math1, do_gamma_cody_i);


    /* Math 2 Group */
    replaceBuiltin("lbeta", do_math2, do_lbeta_i);
    replaceBuiltin("beta", do_math2, do_beta_i);
    replaceBuiltin("lchoose", do_math2, do_lchoose_i);
    replaceBuiltin("choose", do_math2, do_choose_i);

    replaceBuiltin("dchisq", do_math2, do_dchisq_i);
    replaceBuiltin("pchisq", do_math2, do_pchisq_i);
    replaceBuiltin("qchisq", do_math2, do_qchisq_i);

    replaceBuiltin("dexp", do_math2, do_dexp_i);
    replaceBuiltin("pexp", do_math2, do_pexp_i);
    replaceBuiltin("qexp", do_math2, do_qexp_i);

    replaceBuiltin("dgeom", do_math2, do_dgeom_i);
    replaceBuiltin("pgeom", do_math2, do_pgeom_i);
    replaceBuiltin("qgeom", do_math2, do_qgeom_i);

    replaceBuiltin("dpois", do_math2, do_dpois_i);
    replaceBuiltin("ppois", do_math2, do_ppois_i);
    replaceBuiltin("qpois", do_math2, do_qpois_i);

    replaceBuiltin("dt", do_math2, do_dt_i);
    replaceBuiltin("pt", do_math2, do_pt_i);
    replaceBuiltin("qt", do_math2, do_qt_i);

#ifdef NOT_THREAD_SAFE
    replaceBuiltin("dsignrank", do_math2, do_dsignrank_i);
    replaceBuiltin("psignrank", do_math2, do_psignrank_i);
    replaceBuiltin("qsignrank", do_math2, do_qsignrank_i);

    replaceBuiltin("bessel_j", do_math2, do_bessel_j_i);
    replaceBuiltin("bessel_y", do_math2, do_bessel_y_i);
#endif
    replaceBuiltin("psigamma", do_math2, do_psigamma_i);


    /* Math 3 Group */
    replaceBuiltin("dbeta", do_math3, do_dbeta_i);
    replaceBuiltin("pbeta", do_math3, do_pbeta_i);
    replaceBuiltin("qbeta", do_math3, do_qbeta_i);

    replaceBuiltin("dbinom", do_math3, do_dbinom_i);
    replaceBuiltin("pbinom", do_math3, do_pbinom_i);
    replaceBuiltin("qbinom", do_math3, do_qbinom_i);

    replaceBuiltin("dcauchy", do_math3, do_dcauchy_i);
    replaceBuiltin("pcauchy", do_math3, do_pcauchy_i);
    replaceBuiltin("qcauchy", do_math3, do_qcauchy_i);

    replaceBuiltin("df", do_math3, do_df_i);
    replaceBuiltin("pf", do_math3, do_pf_i);
    replaceBuiltin("qf", do_math3, do_qf_i);

    replaceBuiltin("dgamma", do_math3, do_dgamma_i);
    replaceBuiltin("pgamma", do_math3, do_pgamma_i);
    replaceBuiltin("qgamma", do_math3, do_qgamma_i);

    replaceBuiltin("dlnorm", do_math3, do_dlnorm_i);
    replaceBuiltin("plnorm", do_math3, do_plnorm_i);
    replaceBuiltin("qlnorm", do_math3, do_qlnorm_i);

    replaceBuiltin("dlogis", do_math3, do_dlogis_i);
    replaceBuiltin("plogis", do_math3, do_plogis_i);
    replaceBuiltin("qlogis", do_math3, do_qlogis_i);

    replaceBuiltin("dnbinom", do_math3, do_dnbinom_i);
    replaceBuiltin("pnbinom", do_math3, do_pnbinom_i);
    replaceBuiltin("qnbinom", do_math3, do_qnbinom_i);

    replaceBuiltin("dnorm", do_math3, do_dnorm_i);
    replaceBuiltin("pnorm", do_math3, do_pnorm_i);
    replaceBuiltin("qnorm", do_math3, do_qnorm_i);

    replaceBuiltin("dunif", do_math3, do_dunif_i);
    replaceBuiltin("punif", do_math3, do_punif_i);
    replaceBuiltin("qunif", do_math3, do_qunif_i);

    replaceBuiltin("dweibull", do_math3, do_dweibull_i);
    replaceBuiltin("pweibull", do_math3, do_pweibull_i);
    replaceBuiltin("qweibull", do_math3, do_qweibull_i);

    replaceBuiltin("dnchisq", do_math3, do_dnchisq_i);
    replaceBuiltin("pnchisq", do_math3, do_pnchisq_i);
    replaceBuiltin("qnchisq", do_math3, do_qnchisq_i);

    replaceBuiltin("dnt", do_math3, do_dnt_i);
    replaceBuiltin("pnt", do_math3, do_pnt_i);
    replaceBuiltin("qnt", do_math3, do_qnt_i);

#ifdef NOT_THREAD_SAFE
    replaceBuiltin("dwilcox", do_math3, do_dwilcox_i);
    replaceBuiltin("pwilcox", do_math3, do_pwilcox_i);
    replaceBuiltin("qwilcox", do_math3, do_qwilcox_i);

    replaceBuiltin("bessel_i", do_math3, do_bessel_i_i);
    replaceBuiltin("bessel_k", do_math3, do_bessel_k_i);
#endif


    /* Math 4 Group */
    replaceBuiltin("dhyper", do_math4, do_dhyper_i);
    replaceBuiltin("phyper", do_math4, do_phyper_i);
    replaceBuiltin("qhyper", do_math4, do_qhyper_i);

    replaceBuiltin("dnbeta", do_math4, do_dnbeta_i);
    replaceBuiltin("pnbeta", do_math4, do_pnbeta_i);
    replaceBuiltin("qnbeta", do_math4, do_qnbeta_i);

    replaceBuiltin("dnf", do_math4, do_dnf_i);
    replaceBuiltin("pnf", do_math4, do_pnf_i);
    replaceBuiltin("qnf", do_math4, do_qnf_i);

#ifdef UNIMP
    replaceBuiltin("dtukey", do_math4, do_dtukey_i);
#endif
    replaceBuiltin("ptukey", do_math4, do_ptukey_i);
    replaceBuiltin("qtukey", do_math4, do_qtukey_i);

    return R_NilValue;
}
